using System;
using System.Collections.Generic;
using System.IO;

namespace XELF.Checksum {

	public static class Crc32 {
		private static uint[] Table = new uint[256];

		static Crc32() {
			for (int n = 0; n < Table.Length; n++) {
				uint c = (uint)n;
				for (int k = 0; k < 8; k++) {
					if ((c & (0x00000001)) > 0)
						c = 0xEDB88320 ^ (c >> 1);
					else
						c = c >> 1;
				}
				Table[n] = c;
			}
		}
		private static uint Compute(uint Crc, byte[] Buffer, int Index, int Length) {
			uint c = Crc;

			for (int n = Index, end = Index + Length; n < end; n++) {
				c = Table[(c ^ Buffer[n]) & 0xFF] ^ (c >> 8);
			}

			return c;
		}
		public static uint Get(byte[] Buffer, int Index, int Length) {
			return Compute(0xFFFFFFFF, Buffer, Index, Length) ^ 0xFFFFFFFF;
		}
		public static uint Update(uint Crc, byte[] Buffer, int Index, int Length) {
			return Compute(Crc ^ 0xFFFFFFFF, Buffer, Index, Length) ^ 0xFFFFFFFF;
		}
	}

	public class Adler32 {
		private const int Base = 65521;

		uint value;

		public static uint Get(byte[] Buffer) {
			Adler32 adler = new Adler32();
			adler.Update(Buffer, 0, Buffer.Length);
			return adler.Current;
		}

		public void Reset() { value = 1; }

		public Adler32() {
			Reset();
		}

		public void Update(byte[] Buffer, int Index, int Length) {
			uint s1 = value & 0xFFFF;
			uint s2 = value >> 16;

			for (int i = Index, end = Index + Length; i < end; i++) {
				s1 = (s1 + (uint)Buffer[i]) % Base;
				s2 = (s2 + s1) % Base;
			}

			value = (s2 << 16) + s1;
		}
		public uint Current {
			get { return value; }
		}
	}

}

namespace XELF.Png {

	public enum ColorType : byte {
		Grayscale = 0, // depth: 1,2,4,8,16
		RGB = 2, // depth: 8,16
		PaletteIndex = 3, // depth: 1,2,4,8 + PLTE chunk
		GrayscaleAlpha = 4, // depth: 8,16
		/// <summary>
		/// t@CC[W̏R,G,B,A
		/// </summary>
		RGBA = 6, // depth = 8,16
	}
	public enum InterlaceMethod : byte {
		None = 0,
		Adam7 = 1,
	}

	public struct Tag {
		uint tag;

		public Tag(byte[] Value) {
			if (Value == null)
				throw new ArgumentNullException();
			if (Value.Length != 4)
				throw new ArgumentException();
			tag = (uint)(Value[0] + (Value[1] << 8) + (Value[2] << 16) + (Value[3] << 24));
		}
		public Tag(uint Value) {
			tag = Value;
		}
		public Tag(char A, char B, char C, char D) {
			tag = (uint)(A + (B << 8) + (C << 16) + (D << 24));
		}

		public void Write(BinaryWriter Writer) {
			Writer.Write(tag);
		}
		public override string ToString() {
			return string.Format("{0}{1}{2}{3}",
				(char)(tag & 0xFF),
			 (char)((tag >> 8) & 0xFF),
			 (char)((tag >> 16) & 0xFF),
			 (char)((tag >> 24) & 0xFF));
		}
		public static bool operator ==(Tag X, Tag Y) {
			return X.tag == Y.tag;
		}
		public static bool operator !=(Tag X, Tag Y) {
			return X.tag != Y.tag;
		}
		public override bool Equals(object o) {
			if (o is Tag) {
				return this == (Tag)o;
			}
			return false;
		}
		public override int GetHashCode() {
			return base.GetHashCode();
		}
	}

	public class Png {
		private static byte[] Signature = { 0x89, 0x50, 0x4E, 0x47, 0x0D, 0x0A, 0x1A, 0x0A };

		private readonly IHDRChunk ihdr = new IHDRChunk();
		private readonly IDATChunk idat = new IDATChunk();
		private readonly IENDChunk iend = new IENDChunk();
		private readonly List<UserChunk> userChunks = new List<UserChunk>();

		public IHDRChunk IHDR { get { return ihdr; } }
		public IDATChunk IDAT { get { return idat; } }
		public List<UserChunk> UserChunks { get { return userChunks; } }
		public IENDChunk IEND { get { return iend; } }

		public void SetImage(byte[] ImageData, int Width, int Height) {
			ihdr.Width = Width;
			ihdr.Height = Height;
			ihdr.BitDepth = 8;
			ihdr.ColorType = ColorType.RGBA;
			ihdr.CompressionMethod = 0;
			ihdr.FilterMethod = 0;
			ihdr.InterlaceMethod = InterlaceMethod.None;
			idat.SetContent(ImageData, Width, Height);
		}

		static bool MatchSignature(Reader Reader) {
			byte[] signature = Reader.ReadBytes(Signature.Length);
			for (int i = 0; i < Signature.Length; i++) {
				if (Signature[i] != signature[i])
					return false;
			}
			return true;
		}

		public class Chunks {
			readonly List<UserChunk> items;
			protected Chunks(List<UserChunk> Chunks) {
				items = Chunks;
			}
			public IEnumerable<UserChunk> this[Tag tag] {
				get {
					foreach (UserChunk chunk in items) {
						if (chunk.Tag == tag) yield return chunk;
					}
				}
			}
			public static Chunks From(Reader Reader) {

				List<UserChunk> chunks = new List<UserChunk>();
				for (; ; ) {
					UserChunk chunk = Reader.ReadChunk();
					chunks.Add(chunk);
					if (chunk.Tag == IENDChunk.Type)
						break;
				}

				return new Chunks(chunks);
			}
		}
		Chunks ChunksFrom(Stream Stream) {
			using (Reader reader = new Reader(Stream)) {

				MatchSignature(reader);

				//ihdr
				//idat

				return Chunks.From(reader);
			}

		}

		public Chunks LoadChunks(string Path) {
			using (Stream stream = File.OpenRead(Path)) {
				Chunks chunks = ChunksFrom(stream);
				return chunks;
			}
		}

		public void Save(string Path) {
			using (Stream stream = File.Create(Path)) {
				ToStream().WriteTo(stream);
			}
		}
		MemoryStream ToStream() {
			MemoryStream memoryStream = new MemoryStream();
			using (Writer writer = new Writer(memoryStream)) {
				writer.Write(Signature);
				writer.Write(ihdr);
				writer.Write(idat);
				foreach (IChunk chunk in userChunks) {
					writer.Write(chunk);
				}
				writer.Write(iend);

				memoryStream.Position = 0;
			}
			return memoryStream;
		}
	}

	public class Reader : IDisposable {
		Stream Stream;
		BinaryReader BinaryReader;

		public void Dispose() {

		}
		public Reader(Stream Stream) {
			this.Stream = Stream;
			BinaryReader = new BinaryReader(Stream);
		}
		public UserChunk ReadChunk() {

			int length = ReadInt32();
			byte[] tag = ReadBytes(4);
			byte[] content = ReadBytes(length);
			uint crc32 = ReadUInt32();
			uint sum = Checksum.Crc32.Get(tag, 0, tag.Length);
			sum = Checksum.Crc32.Update(sum, content, 0, content.Length);
			if (crc32 != sum)
				throw new InvalidDataException();

			return new UserChunk(new Tag(tag), content);
		}
		public byte ReadByte() {
			return BinaryReader.ReadByte();
		}
		public uint ReadUInt32() {
			uint value = (uint)(ReadByte() << 24);
			value |= (uint)(ReadByte() << 16);
			value |= (uint)(ReadByte() << 8);
			value |= ReadByte();
			return value;
		}
		public int ReadInt32() {
			return (int)ReadUInt32();
		}
		public byte[] ReadBytes(int Count) {
			return BinaryReader.ReadBytes(Count);
		}
		public Tag ReadTag() {
			return new Tag(BinaryReader.ReadUInt32());
		}
	}

	public class Writer : IDisposable {
		MemoryStream Stream;
		BinaryWriter BinaryWriter;

		public Writer(MemoryStream Stream) {
			this.Stream = Stream;
			this.BinaryWriter = new BinaryWriter(Stream);
		}
		public void Dispose() {
			//BinaryWriter.Close();
		}
		public void Write(IChunk Value) {
			Write(0); // ̒
			Stream.Flush();
			long begin = Stream.Position;
			Value.Tag.Write(BinaryWriter);
			Value.WriteContent(this);
			Stream.Flush();
			long end = Stream.Position;
			Stream.Position = begin - 4;
			Write((int)(end - begin - 4)); // 
			Stream.Position = end;
			Write(Checksum.Crc32.Get(Stream.GetBuffer(), (int)begin, (int)(end - begin)));
		}
		public void Write(byte Value) {
			BinaryWriter.Write(Value);
		}
		public void Write(int Value) {
			BinaryWriter.Write((byte)(Value >> 24));
			BinaryWriter.Write((byte)(Value >> 16));
			BinaryWriter.Write((byte)(Value >> 8));
			BinaryWriter.Write((byte)(Value));
		}
		public void Write(uint Value) {
			BinaryWriter.Write((byte)(Value >> 24));
			BinaryWriter.Write((byte)(Value >> 16));
			BinaryWriter.Write((byte)(Value >> 8));
			BinaryWriter.Write((byte)(Value));
		}

		public void Write(byte[] Value) {
			BinaryWriter.Write(Value);
		}
	}

	public interface IChunk {
		Tag Tag { get; }
		void WriteContent(Writer Writer);
		void ReadContent(Reader Reader);
	}

	public class IHDRChunk : IChunk {
		public int Width;
		public int Height;
		public byte BitDepth;
		public ColorType ColorType;
		public byte CompressionMethod;
		public byte FilterMethod;
		public InterlaceMethod InterlaceMethod;

		private static readonly Tag tag = new Tag('I', 'H', 'D', 'R');

		public Tag Tag { get { return tag; } }
		public void WriteContent(Writer Writer) {
			Writer.Write(Width);
			Writer.Write(Height);
			Writer.Write(BitDepth);
			Writer.Write((byte)ColorType);
			Writer.Write(CompressionMethod);
			Writer.Write(FilterMethod);
			Writer.Write((byte)InterlaceMethod);
		}
		public void ReadContent(Reader Reader) {
			throw new NotImplementedException();
		}
	}

	public class IDATChunk : IChunk {
		public static readonly Tag Type = new Tag('I', 'D', 'A', 'T');
		const int Size = 0xFFFF;

		private byte[] content;
		private int width;
		private int height;

		public byte[] Content { get { return content; } }
		public int Width { get { return width; } }
		public int Height { get { return height; } }
		public void SetContent(byte[] Content, int Width, int Height) {
			content = Content;
			width = Width;
			height = Height;
		}

		public Tag Tag { get { return Type; } }
		public void WriteContent(Writer Writer) {
			Checksum.Adler32 adler = new Checksum.Adler32();
			MemoryStream compession = new MemoryStream();

			int BytesPerPixel = 4; // ݂͌Œ

			uint contentWidthLength = (uint)(Width * BytesPerPixel);
			uint widthLength = contentWidthLength + 1;
			uint dcSize = widthLength * (uint)Height;
			uint rowsPerBlock = Size / widthLength;
			uint blockSize = rowsPerBlock * widthLength;
			uint contentBlockSize = rowsPerBlock * contentWidthLength;
			uint blockCount;
			ushort length;
			uint remainder = dcSize;

			blockCount = (dcSize + (blockSize - 1)) / blockSize;

			// wb_
			compession.WriteByte(0x78);
			compession.WriteByte(0xDA);

			byte[] lineHeader = new byte[] { 0 };

			int y = 0;
			for (uint blocks = 0; blocks < blockCount; blocks++) {
				// 
				length = (ushort)Math.Min(remainder, blockSize);

				byte lastRaw = 0x01;
				byte continueRaw = 0x00;
				compession.WriteByte((length == remainder) ? lastRaw : continueRaw);

				compession.Write(BitConverter.GetBytes(length), 0, 2);

				// 1̕␔
				compession.Write(BitConverter.GetBytes((ushort)~length), 0, 2);

				int end = Math.Min(height, y + (int)rowsPerBlock);
				for (; y < end; y++) {
					adler.Update(lineHeader, 0, 1);
					compession.WriteByte(lineHeader[0]);
					int start = y * (int)contentWidthLength;
					adler.Update(Content, start, (int)contentWidthLength);
					compession.Write(Content, start, (int)contentWidthLength);
				}
				remainder -= blockSize;
			}

			Writer.Write(compession.ToArray());
			Writer.Write(adler.Current);
		}
		public void ReadContent(Reader Reader) {
			throw new NotImplementedException();
		}
	}

	public class IENDChunk : IChunk {
		public static readonly Tag Type = new Tag('I', 'E', 'N', 'D');

		public Tag Tag { get { return Type; } }
		public void WriteContent(Writer Writer) { }
		public void ReadContent(Reader Reader) {
			throw new NotImplementedException();
		}
	}

	public class UserChunk : IChunk {
		private readonly Tag tag;
		public byte[] Content;

		public UserChunk(Tag Tag)
			: this(Tag, null) {
		}
		public UserChunk(Tag Tag, byte[] Content) {
			this.tag = Tag;
			this.Content = Content;
		}

		public Tag Tag { get { return tag; } }
		public void WriteContent(Writer Writer) {
			if (Content != null)
				Writer.Write(Content);
		}
		public void ReadContent(Reader Reader) {
			throw new NotSupportedException();
		}
		public override string ToString() {
			return string.Format("{0} {{{1}}}", Tag, Content == null ? 0 : Content.Length);
		}
	}
}
